/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file slider.c
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "kernel.h"

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "msg.h"
#include "draw.h"
#include "drag.h"

#define SLIDER_COMMENT (1<<0)
#define SLIDER_TEXT (1<<1)
#define SLIDER_UNITS (1<<2)
#define SLIDER_ARROWS (1<<3)
#define SLIDER_CENTRE (1<<4)

#define SLIDER_RETURNPRESSED (0)
#define SLIDER_INCREMENT     (1)
#define SLIDER_DECREMENT     (2)
#define SLIDER_DRAGTO        (3)
#define SLIDER_DRAGEND       (4)
#define SLIDER_KEYPRESSED    (5)

typedef int (* slider_changeval)(void * , int, int);

typedef struct slider {
  struct slider * next;
  canvas c;
  icon i[6]; /* comment, text, units, left arrow, centre, right arrow */
  int x, y, w, step;
  int flags;
  int min, max, value;
  int dp;
  char val[12];
  int col, cutoff, old;
  slider_changeval rout;
} slider;

static slider * s = NULL, * sd;

static int dps[5] = {1, 10, 100, 1000, 10000};

void slider_redraw(canvas c, int x0, int y0, int x1, int y1);

static int absi(int n)
{
  if(n < 0) return - n;
  return n;
}

void slider_setvalue(slider * ss, int val)
{
  if(val == ss -> old) return;

  ss -> old = ss -> value = val;
  if(ss -> flags & SLIDER_TEXT) {
    if(ss -> dp == 0) {
      sprintf(ss -> val, "%d", val);
    } else {
      char c[12];
      if(val < 0) {
	sprintf(c, "-%%d.%%-.%dd", ss -> dp);
      } else {
	sprintf(c, "%%d.%%-.%dd", ss -> dp);
      }
      sprintf(ss -> val, c, absi(val) / dps[ss -> dp], absi(val)%dps[ss -> dp]);
    }
    if(ss -> i[1]) icon_update(ss -> i[1]);
  }
  if(ss -> i[4]) canvas_update(ss -> c, ss -> x - 4, ss -> y - 4, ss -> x + ss -> w + 8, ss -> y + 20, slider_redraw);
}

static void _slider_inc(slider * s)
{
  s -> value += s -> step;
  if(s -> value > s -> max) s -> value = s -> max;
  if(s -> rout) {
    if(s -> rout((void *)s, SLIDER_INCREMENT, s -> value)) return;
  }
  slider_setvalue(s, s -> value);
}

static void _slider_dec(slider * s)
{
  s -> value -= s -> step;
  if(s -> value < s -> min) s -> value = s -> min;
  if(s -> rout) {
    if(s -> rout((void *)s, SLIDER_DECREMENT, s -> value)) return;
  }
  slider_setvalue(s, s -> value);
}

static void _slider_enddrag(main_mousestr *m)
{
  if(sd -> rout) {
    if(sd -> rout((void *)sd, SLIDER_DRAGEND, sd -> value)) return;
  }
  slider_setvalue(sd, sd -> value);
}

static void _slider_durdrag(int x, int y)
{
  int c = sd -> min + ((x - sd -> x) * (sd -> max - sd -> min)) / sd -> w;

  if(c < sd -> min) c = sd -> min;
  if(c > sd -> max) c = sd -> max;
  sd -> value = c;
  if(sd -> rout) {
    if(sd -> rout((void *)sd, SLIDER_DRAGTO, sd -> value)) return;
  }
  slider_setvalue(sd, sd -> value);
}

static void _slider_drag(canvas c, icon i, canvas_mousestr * m)
{
  slider * ss = s;
  while(ss) {
    if(ss -> i[4] == i) {
      switch(m -> b) {
      case 1 : /* Adjust */
	if(m -> x < ss -> cutoff + ss -> x) {
	  _slider_inc(ss);
	} else {
	  _slider_dec(ss);
	}
	break;
      case 4 : /* Select */
	if(m -> x > ss -> cutoff + ss -> x) {
	  _slider_inc(ss);
	} else {
	  _slider_dec(ss);
	}
	break;
      case 16 : /* Drag Adjust */
      case 64 : /* Drag Select */
	drag_htrench(ss -> c, ss -> x, ss -> y + 8, ss -> w, _slider_enddrag, _slider_durdrag);
	sd = ss;
	break;
      }
      return;
    }
    ss = ss -> next;
  }
}

void slider_redraw(canvas c, int x0, int y0, int x1, int y1)
{
  slider * ss = s;
  int w;

  while(ss) {
    if(c == ss -> c) {
      if(!((x0 > ss -> x + ss -> w) || (x1 < ss -> x) || (y0 > ss -> y + 24) || (y1 <= ss -> y))) {
	ss -> cutoff = w = ((ss -> w * (ss -> value - ss -> min)) / (ss -> max - ss -> min));
	draw_move(ss -> x + ss -> w, ss -> y);
	if(w != ss -> w) {
	  draw_gcol(0);
	  draw_plot(0x61, - ss -> w + w, 16);
	  draw_moveto(0, - 16);
	}
	if(w != 0) {
	  draw_gcol(ss -> col);
	  draw_plot(0x61, - w, 16);
	}
      }
    }
    ss = ss -> next;
  }
}

static void _slider_idci(canvas c, icon i, canvas_mousestr * m)
{
  int d = 1;
  slider * ss = s;

  if(m -> b == 1) d = 0;

  while(ss) {
    if(ss -> i[3] == i) {
      d ^= 1;
    }
    if((ss -> i[5] == i) || (ss -> i[3] == i)) {
      if(d) {
	_slider_inc(ss);
      } else {
	_slider_dec(ss);
      }
      return;
    }
    ss = ss -> next;
  }
}

int _slider_getvalue(slider * ss)
{
  char str[12], * j, * k/*, l*/;
  int m, o = 1;

  if(ss -> dp == 0) return atoi(ss -> val);
  j = ss -> val;
  k = str;
  if(* j == '-') {
    o = -1;
    j ++ ;
  }
  //l = 0;
  while(isdigit(* j)) * k ++ = * j ++ ;
  if(* j == '.') {
    m = ss -> dp;
    j ++ ;
    while(isdigit(* j) && m) {
      * k ++ = * j ++ ;
      m -- ;
    }
    return atoi(str) * dps[m] * o;
  } else {
    * k = 0;
    return atoi(str) * dps[ss -> dp] * o;
  }
}

static void _slider_key(canvas c, icon i, canvas_keystr * k)
{
  slider * ss = s;
  int val;

  while((ss) && (i != ss -> i[1])) {
    ss = ss -> next;
  }
  if(k -> chcode > 0x100) main_processkey(k -> chcode);
  if(!ss) return;
  val = _slider_getvalue(ss);
  ss -> old = ss -> min - 1;
  if(val > ss -> max) val = ss -> max;
  if(val < ss -> min) val = ss -> min;
  if(k -> chcode == 27) val = ss -> value;
  switch(k -> chcode) {
  case 27 :
  case 10 :
  case 13 :
  case 0x18a :
  case 0x19a :
  case 0x1aa :
  case 0x1ba :
  case 0x18e :
  case 0x19e :
  case 0x1ae :
  case 0x1be :
  case 0x18f :
  case 0x19f :
  case 0x1af :
  case 0x1bf :
    if(ss -> rout) {
      if(ss -> rout((void *)ss, SLIDER_RETURNPRESSED, val)) return;
    }
    slider_setvalue(ss, val);
    return;
  }
  if(ss -> rout) {
    ss -> rout((void *)ss, SLIDER_KEYPRESSED, val);
  }
}

slider * slider_create(canvas c, int x, int y, int w, int flags, char * comment, char * units, int value, int min, int max, int step, int dp, int tw, int uw, slider_changeval r)
{
  int centre_x = x, centre_w = w, i;
  slider * ss = rmalloc(sizeof(slider));
  icon_data d;

  if(ss == 0) return 0;

  ss -> next = s;
  s = ss;
  ss -> rout = r;
  ss -> col = COLOUR_DARKGREEN;
  ss -> c = c;
  ss -> min = min;
  ss -> max = max;
  ss -> value = value;
  ss -> step = step;
  ss -> dp = dp;
  ss -> flags = flags;
  ss -> old = ss -> min - 1;

  for(i = 0; i < 6; i ++) ss -> i[i] = 0;

  if(flags & SLIDER_TEXT) {
    slider_setvalue(ss, value);
  }

  if(flags & SLIDER_COMMENT) {
    ss -> i[0] = icon_createrjtext(c, x - 4, y, comment);
  }

  if(flags & SLIDER_TEXT) {
    if(tw == 0) tw = main_strwidth(ss -> val) + main_strwidth("-") + 16;
    if(ss -> step) {
      if(min < 0) {
	ss -> i[1] = icon_createwriteable(c, centre_x, y, ss -> val, "pptr_write,4,4;a0-9.\\-;kartn", 12, tw);
      } else {
	ss -> i[1] = icon_createwriteable(c, centre_x, y, ss -> val, "pptr_write,4,4;a0-9.;kartn", 12, tw);
      }   } else {
	ss -> i[1] = icon_createfixedinfofield(c, centre_x, y, ss -> val, tw);
      }
    centre_x += tw;
    centre_w -= tw;
  }

  if(flags & SLIDER_UNITS) {
    if(uw == 0) uw = main_strwidth(units) + 16;
    ss -> i[2] = icon_createtext(c, centre_x, y, units);
    centre_x += uw;
    centre_w -= uw;
  }

  if(flags & SLIDER_ARROWS) {
    if(get_3dstatus()) {
      d.indirecttext.text = "";
      d.indirecttext.valid = "sleft,pleft;r5,2";
      d.indirecttext.length = 12;
      ss -> i[3] = icon_create(c, centre_x, y + 4, centre_x + 32, y + 44, 0x1700211b, & d);
    } else {
      d.text[0] = 136;
      d.text[1] = 0;
      ss -> i[3] = icon_create(c, centre_x, y + 4, centre_x + 32, y + 44, 0x3700203d, & d);
    }
    centre_x += 36;
    centre_w -= 72;
  }

  if(flags & SLIDER_CENTRE) {
    if(get_3dstatus()) {
      d.indirecttext.text = "";
      d.indirecttext.valid = "r1";
      d.indirecttext.length = 1;
      ss -> i[4] = icon_create(c, centre_x, y + 4, centre_x + centre_w, y + 44, 0x17006105, & d);
    } else {
      d.text[0] = 0;
      ss -> i[4] = icon_create(c, centre_x, y + 4, centre_x + centre_w, y + 44, 0x17006005, & d);
    }
    ss -> x = centre_x + 12;
    ss -> y = y + 16;
    ss -> w = centre_w - 25;
    centre_x += centre_w + 4;
  }

  if(flags & SLIDER_ARROWS) {
    if(get_3dstatus()) {
      d.indirecttext.text = "";
      d.indirecttext.valid = "sright,pright;r5,2";
      d.indirecttext.length = 12;
      ss -> i[5] = icon_create(c, centre_x, y, centre_x + 32, y + 52, 0x1700211b, & d);
    } else {
      d.text[0] = 137;
      d.text[1] = 0;
      ss -> i[5] = icon_create(c, centre_x, y + 8, centre_x + 32, y + 52 - 8, 0x3700203d, & d);
    }
  }

  if(ss -> i[4]) icon_addhandler(ss -> i[4], ICON_CLICKHANDLER, _slider_drag);
  if(ss -> i[3]) icon_addhandler(ss -> i[3], ICON_CLICKHANDLER, _slider_idci);
  if(ss -> i[5]) icon_addhandler(ss -> i[5], ICON_CLICKHANDLER, _slider_idci);
  if(ss -> i[1]) icon_addhandler(ss -> i[1], ICON_KEYHANDLER, _slider_key);

  return ss;
}

void slider_setcolour(slider * s, int c)
{
  s -> col = c;
}
